import {DataSource} from '../backend/data-source';
import {validation as v} from '.';
import {NamedSet} from './common';

export type EndpointQueue = [Endpoint, Record<string, unknown> | undefined][];

export const serializeEndpointQueue = (endpointQueue: EndpointQueue) => {
	return endpointQueue
		.map(([endpoint, param]) => {
			if (typeof param === 'undefined') {
				return `/${endpoint.name}`;
			}

			return [
				endpoint.name,
				...Array.from(endpoint.params).map((s) => param[s] ?? '_')
			]
				.map((s) => `/${s}`)
				.join('');
		})
		.join('')
};

export const parseToEndpointQueue = (urlWithoutBase: string, endpoints: NamedSet<Endpoint>) => {
	// why we don't get the url without query params as parameter, because we might need the query params in the future
	const [urlWithoutQueryParams] = urlWithoutBase.split('?');
	const fragments = urlWithoutQueryParams.split('/').filter((s) => s.trim().length > 0);

	return fragments.reduce(
		(theEndpointQueueRaw, s) => {
			const theEndpointQueue = theEndpointQueueRaw as EndpointQueue;
			const [lastEndpoint, lastEndpointParams] = theEndpointQueue.at(-1) ?? [];
			const endpoint = endpoints.get(s);
			if (typeof endpoint !== 'undefined') {
				if (typeof lastEndpoint === 'undefined') {
					return [
						...theEndpointQueue,
						[endpoint, {}]
					];
				}
			}

			if (typeof lastEndpoint === 'undefined') {
				throw new Error(`Invalid URL: ${urlWithoutBase}`);
			}
			const lastEndpointParamsOrdering = Array.from(lastEndpoint.params);
			const lastEndpointParamsOrderingLength = lastEndpointParamsOrdering.length;
			if (lastEndpointParamsOrderingLength > 0) {
				if (typeof lastEndpointParams === 'undefined') {
					return [
						...theEndpointQueue.slice(0, -1),
						[lastEndpoint, {
							[lastEndpointParamsOrdering[0]]: s
						}]
					];
				}

				const nextIndex = Object.keys(lastEndpointParams).length;
				if (nextIndex === lastEndpointParamsOrderingLength) {
					throw new Error(`Invalid URL: ${urlWithoutBase}`);
				}

				return [
					...theEndpointQueue.slice(0, -1),
					[lastEndpoint, {
						...lastEndpointParams,
						[lastEndpointParamsOrdering[nextIndex]]: s
					}]
				];
			}

			throw new Error(`Invalid URL: ${urlWithoutBase}`);
		},
		[] as unknown
	) as EndpointQueue;
};

interface BaseEndpointState {
	operations: unknown;
	params: unknown;
}

type OpValueType = undefined | boolean | Record<string, boolean> | readonly string[];

export interface Endpoint<
	Name extends string = string,
	Schema extends v.BaseSchema = v.BaseSchema,
	State extends BaseEndpointState = BaseEndpointState
> {
	name: Name;
	dataSource?: DataSource;
	schema: Schema;
	params: Set<string>;
	operations: Set<string>;
	can<OpName extends string = string, OpValue extends OpValueType = OpValueType>(
		op: OpName,
		value?: OpValue
	): Endpoint<
		Name,
		Schema,
		{
			operations: State['operations'] extends string[] ? readonly [...State['operations'], OpName] : [OpName];
			params: State['params']
		}
	>;
	param<ParamName extends string = string>(name: ParamName): Endpoint<
		Name,
		Schema,
		{
			operations: State['operations'];
			params: State['params'] extends string[] ? readonly [...State['params'], ParamName]: [ParamName];
		}
	>
}

export type GetEndpointParams<T extends Endpoint> = T extends Endpoint<any, infer R> ? (
	R extends { params: Record<number, any> } ? R['params'][number] : never
) : never;

interface EndpointParams<Name extends string = string, Schema extends v.BaseSchema = v.BaseSchema> {
	name: Name;
	schema: Schema;
	dataSource?: DataSource;
}

class EndpointInstance<
	Params extends EndpointParams,
	State extends BaseEndpointState
> implements Endpoint<Params['name'], Params['schema'], State> {
	readonly name: Params['name'];
	readonly dataSource: Params['dataSource'];
	readonly operations: Set<string>;
	readonly params: Set<string>;
	readonly schema: Params['schema'];

	constructor(params: Params) {
		this.name = params.name;
		this.schema = params.schema;
		this.operations = new Set<string>();
		this.params = new Set<string>();
		this.dataSource = params.dataSource;
	}

	can<OpName extends string = string, OpValue extends OpValueType = OpValueType>(
		op: OpName,
		value = true as OpValue
	) {
		if (value) {
			this.operations.add(op);
		} else {
			// todo remove operation at type level
			this.operations.delete(op);
		}

		return this;
	}

	param<ParamName extends string = string>(
		name: ParamName
	) {
		this.params.add(name);

		return this;
	}
}

export const endpoint = <Params extends EndpointParams>(params: Params): Endpoint<
	Params['name'],
	Params['schema'],
	{
		operations: [];
		params: [];
	}
> => {
	return new EndpointInstance(params);
};


export type EndpointOperations<T extends Endpoint> = T extends Endpoint<string, v.BaseSchema, infer R> ? (
	R extends BaseEndpointState
		? R['operations'] extends []
			? R['operations'] extends readonly string[]
				? R['operations'][number]
				: string
			: string
		: never
) : never;
