import {ErrorStatusCode, isErrorStatusCode, StatusCode} from './status-codes';

type FetchResponse = Awaited<ReturnType<typeof fetch>>;

export interface Response<B extends unknown = unknown> {
	statusCode: number;
	statusMessage: string;
	body?: B;
}

export interface ErrorResponse<B extends unknown = unknown> extends Error, Response<B> {}

export interface HttpResponseConstructor<R extends Response> {
	new (...args: any[]): R;
	fromFetchResponse(response: FetchResponse): R;
}

export interface HttpResponseErrorConstructor<R extends Response> extends HttpResponseConstructor<R> {
	new (message?: string, options?: ErrorOptions): R;
}

export interface HttpSuccessResponseConstructor<R extends Response> extends HttpResponseConstructor<R> {
	new (response: Partial<Omit<Response, 'statusCode'>>): R;
}

export interface HttpErrorOptions<B extends unknown = unknown> extends ErrorOptions {
	body?: B;
}

export const HttpResponse = <
	B extends unknown = unknown,
	T extends StatusCode = StatusCode,
	R extends Response = T extends ErrorStatusCode ? ErrorResponse<B> : Response<B>,
>(statusCode: T): T extends ErrorStatusCode ? HttpResponseErrorConstructor<R> : HttpSuccessResponseConstructor<R> => {
	if (isErrorStatusCode(statusCode)) {
		return class HttpErrorResponse extends Error implements ErrorResponse<B> {
			readonly statusMessage: string;
			readonly statusCode: T;
			readonly body?: B;

			constructor(message?: string, options?: HttpErrorOptions<B>) {
				super(message, options);
				this.name = this.statusMessage = message ?? '';
				this.statusCode = statusCode;
				this.cause = options?.cause;
				this.body = options?.body;
			}
		} as unknown as HttpResponseErrorConstructor<R>;
	}

	return class HttpSuccessResponse implements Response<B> {
		readonly statusMessage: string;
		readonly statusCode: T;
		readonly body?: B;
		constructor(params: Partial<Omit<Response<B>, 'statusCode'>>) {
			this.statusCode = statusCode;
			this.statusMessage = params.statusMessage ?? '';
			this.body = params.body;
		}

		static fromFetchResponse(response: FetchResponse) {
			return {
				statusCode: response.status,
				statusMessage: response.statusText,
				deserialize: async () => {
					if (response.status !== statusCode) {
						throw new Error(`Status codes do not match: ${response.status} !== ${statusCode}`);
					}

					const contentType = response.headers.get('Content-Type');
					// TODO properly parse media type
					if (contentType === 'application/json') {
						return await response.json();
					}

					const buffer = await response.arrayBuffer();
					return buffer;
					// TODO deserialize buffer
				},
			};
		}
	} as unknown as HttpSuccessResponseConstructor<R>;
};
