import http from 'http';
import {ErrorResponse, parseToEndpointQueue, ServiceParams, statusCodes} from '@modal-sh/yasumi';
import {
	Backend as BaseBackend,
	Server,
	ServerRequestContext,
	ServerResponseContext,
	ServerParams,
} from '@modal-sh/yasumi/backend';

declare module '@modal-sh/yasumi/backend' {
	interface ServerRequestContext extends http.IncomingMessage {}

	interface ServerResponseContext extends http.ServerResponse {}
}

class ServerInstance<Backend extends BaseBackend> implements Server<Backend> {
	readonly backend: Backend;
	private serverInternal?: http.Server;

	constructor(params: ServerParams<Backend>) {
		this.backend = params.backend;
	}

	private readonly requestListener = async (req: ServerRequestContext, res: ServerResponseContext) => {
		// const endpoints = this.backend.app.endpoints;
		if (typeof req.method === 'undefined') {
			res.writeHead(statusCodes.HTTP_STATUS_BAD_REQUEST, {});
			res.end();
			return;
		}
		if (typeof req.url === 'undefined') {
			res.writeHead(statusCodes.HTTP_STATUS_BAD_REQUEST, {});
			res.end();
			return;
		}

		console.log(req.method, req.url);

		const endpoints = parseToEndpointQueue(req.url, this.backend.app.endpoints);
		const [endpoint, endpointParams] = endpoints.at(-1) ?? [];

		const appOperations = Array.from(this.backend.app.operations.values())
		const foundAppOperation = appOperations
			.find((op) => op.method === req.method?.toUpperCase());

		if (typeof foundAppOperation === 'undefined') {
			res.writeHead(statusCodes.HTTP_STATUS_METHOD_NOT_ALLOWED, {
				'Allow': appOperations.map((op) => op.method).join(',')
			});
			res.end();
			return;
		}

		if (!endpoint?.operations?.has(foundAppOperation.name)) {
			const endpointOperations = Array.from(endpoint?.operations ?? []);
			res.writeHead(statusCodes.HTTP_STATUS_METHOD_NOT_ALLOWED, {
				'Allow': endpointOperations
					.map((a) => appOperations.find((aa) => aa.name === a)?.method)
					.join(',')
			});
			res.end();
			return;
		}

		const implementation = this.backend.implementations.get(foundAppOperation.name);
		if (typeof implementation === 'undefined') {
			res.writeHead(statusCodes.HTTP_STATUS_NOT_IMPLEMENTED);
			res.end();
			return;
		}

		if (typeof endpoint === 'undefined') {
			res.writeHead(statusCodes.HTTP_STATUS_NOT_IMPLEMENTED);
			res.end();
			return;
		}

		const [, search] = req.url.split('?');

		// TODO get content negotiation params

		// TODO add flag on implementation context if CQRS should be enabled

		try {
			const responseSpec = await implementation({
				endpoint,
				params: endpointParams ?? {},
				query: typeof search !== 'undefined' ? new URLSearchParams(search) : undefined,
				dataSource: this.backend.dataSource,
			});

			if (typeof responseSpec === 'undefined') {
				res.writeHead(statusCodes.HTTP_STATUS_UNPROCESSABLE_ENTITY, {});
				res.end();
				return;
			}

			// TODO serialize using content-negotiation params
			const bodyToSerialize = responseSpec.body;

			res.statusMessage = responseSpec.statusMessage; // TODO add default status message per status code
			res.writeHead(responseSpec.statusCode, {});
			res.end();
		} catch (errorResponseSpecRaw) {
			const responseSpec = errorResponseSpecRaw as ErrorResponse;
			res.statusMessage = responseSpec.statusMessage; // TODO add default status message per status code
			res.writeHead(responseSpec.statusCode, {});
			res.end();
		}
	};

	serve(params: ServiceParams) {
		return new Promise<void>((resolve) => {
			this.serverInternal = new http.Server(this.requestListener);
			this.serverInternal.listen(params.port ?? 80, params.host ?? '0.0.0.0', undefined, resolve);
		});
	}

	close() {
		return new Promise<void>((resolve, reject) => {
			if (typeof this.serverInternal === 'undefined') {
				resolve();
				return;
			}
			this.serverInternal.close((err) => {
				if (err) {
					reject(err);
					return;
				}
				resolve();
			});
		});
	}
}

export const server = <Backend extends BaseBackend>(params: ServerParams<Backend>): Server<Backend> => {
	return new ServerInstance(params);
};
