import {
	App as BaseApp,
	AVAILABLE_EXTENSION_METHODS,
	Endpoint, EndpointQueue,
	GetEndpointParams,
	Operation, serializeEndpointQueue,
	ServiceParams,
} from '@modal-sh/yasumi';
import {Client, ClientParams, ClientConnection} from '@modal-sh/yasumi/client';

declare module '@modal-sh/yasumi/client' {
	interface ClientConnection {
		host: string;
		port: number;
		basePath: string;
	}
}

const DEFAULT_HOST = '0.0.0.0' as const;

const DEFAULT_PORT = 80 as const;

const DEFAULT_METHOD = 'GET' as const;

const EXTENSION_METHOD_EFFECTIVE_METHOD = 'POST' as const;

class ClientInstance<App extends BaseApp> implements Client<App> {
	readonly app: App;
	private readonly fetchFn: typeof fetch;
	private connection?: ServiceParams;
	private endpointQueue = [] as EndpointQueue;

	constructor(params: ClientParams<App>) {
		this.app = params.app;
		this.fetchFn = params.fetch ?? fetch;
	}

	async connect(params: ServiceParams): Promise<ClientConnection> {
		const connection = {
			host: params.host ?? DEFAULT_HOST,
			port: params.port ?? DEFAULT_PORT,
			basePath: params.basePath ?? '',
		};

		this.connection = connection;

		return connection;
	}

	async disconnect() {
		// noop
	}

	at<TheEndpoint extends Endpoint = Endpoint>(endpoint: TheEndpoint, params?: Record<GetEndpointParams<TheEndpoint>, unknown>) {
		if (Array.isArray(this.endpointQueue)) {
			this.endpointQueue?.push([endpoint, params]);
		}
		return this;
	}

	makeRequest(operation: Operation) {
		const baseUrlFragments = [
			this.connection?.host ?? DEFAULT_HOST
		];

		const thePort = (this.connection?.port ?? DEFAULT_PORT);
		if (thePort !== DEFAULT_PORT) {
			baseUrlFragments.push(thePort.toString());
		}

		// TODO how to set to https?
		const scheme = 'http';
		// todo need a way to decode url back to endpoint queue
		const urlString = serializeEndpointQueue(this.endpointQueue);
		this.endpointQueue = [];

		const url = new URL(
			this.connection?.basePath ? `${this.connection.basePath}${urlString}` : urlString,
			`${scheme}://${baseUrlFragments.join(':')}`
		);
		if (typeof operation.searchParams !== 'undefined') {
			url.search = operation.searchParams.toString();
		}
		const rawEffectiveMethod = (operation.method ?? DEFAULT_METHOD).toUpperCase();
		const finalEffectiveMethod = (AVAILABLE_EXTENSION_METHODS as unknown as string[]).includes(rawEffectiveMethod)
			? EXTENSION_METHOD_EFFECTIVE_METHOD
			: rawEffectiveMethod;

		if (typeof operation.body !== 'undefined') {
			return this.fetchFn(
				url,
				{
					method: finalEffectiveMethod,
					body: operation.body as string,
					// TODO inject headers
				},
			);
		}

		return this.fetchFn(
			url,
			{
				method: finalEffectiveMethod,
				// TODO inject headers
			},
		);
	}
}

export const client = <App extends BaseApp>(params: ClientParams<App>): Client<App> => {
	return new ClientInstance(params);
};
