import {Recipe, endpoint, validation as v} from '@modal-sh/yasumi';
import {backend, DataSource} from '@modal-sh/yasumi/backend';
import * as fetchOperation from './implementation/fetch';
import * as createOperation from './implementation/create';
import * as emplaceOperation from './implementation/emplace';
import * as patchDeltaOperation from './implementation/patch-delta';
import * as patchMergeOperation from './implementation/patch-merge';
import * as queryOperation from './implementation/query';
import * as deleteOperation from './implementation/delete';

interface AddResourceRecipeParams {
	endpointName: string;
	dataSource?: DataSource;
}

export const addResourceRecipe = (params: AddResourceRecipeParams): Recipe => (a) => {
	const operations = {
		fetch: fetchOperation.operation,
		create: createOperation.operation,
		emplace: emplaceOperation.operation,
		patchMerge: patchMergeOperation.operation,
		patchDelta: patchDeltaOperation.operation,
		query: queryOperation.operation,
		delete: deleteOperation.operation,
	};

	const theEndpoint = endpoint({
		name: params.endpointName,
		schema: v.object({
			username: v.string(),
		}),
	})
		.param('resourceId')
		.can(fetchOperation.name)
		.can(createOperation.name)
		.can(emplaceOperation.name)
		.can(patchMergeOperation.name)
		.can(patchDeltaOperation.name)
		.can(queryOperation.name)
		.can(deleteOperation.name);

	const enhancedApp = a.app
		.operation(operations.fetch)
		.operation(operations.create)
		.operation(operations.emplace)
		.operation(operations.patchMerge)
		.operation(operations.patchDelta)
		.operation(operations.query)
		.operation(operations.delete)
		.endpoint(theEndpoint);

	const theBackend = a.backend ?? backend({
		app: enhancedApp,
		dataSource: params.dataSource,
	});

	theBackend
		.implementOperation(fetchOperation.name, fetchOperation.implementation)
		.implementOperation(createOperation.name, createOperation.implementation)
		.implementOperation(deleteOperation.name, deleteOperation.implementation)
		.implementOperation(queryOperation.name, queryOperation.implementation)
		.implementOperation(patchDeltaOperation.name, patchDeltaOperation.implementation)
		.implementOperation(patchMergeOperation.name, patchMergeOperation.implementation)
		.implementOperation(emplaceOperation.name, emplaceOperation.implementation);

	return {
		operations,
		app: enhancedApp,
		backend: theBackend,
		endpoints: {
			[params.endpointName]: theEndpoint,
		},
	};
};
